LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;

LIBRARY work;
USE work.jddr_pkg.all;

ENTITY jddr_controller IS
   GENERIC(
     STARTUPDELAY : natural := 33334;
     EXTRA_WRITE_DELAY : natural := 1;
     ELMR_VALUE : STD_LOGIC_VECTOR(11 downto 0) := "000000000000";
     LMR_VALUE1 : STD_LOGIC_VECTOR(11 downto 0) := "000101100011";
     LMR_VALUE2 : STD_LOGIC_VECTOR(11 downto 0) := "000001100011"
   );
   PORT(
     -- Data transfer interface
     data_available : OUT STD_LOGIC;
     data_wanted : OUT STD_LOGIC;
     -- DDR Datapath interface
     doing_wr : OUT STD_LOGIC;
     doing_rd : OUT STD_LOGIC;
     -- DDR Signals
     ddr_clken : OUT STD_LOGIC_VECTOR(1 downto 0);
     ddr_csn : OUT STD_LOGIC_VECTOR(1 downto 0);
     ddr_rasn : OUT STD_LOGIC;
     ddr_casn : OUT STD_LOGIC;
     ddr_wen : OUT STD_LOGIC;
     ddr_address : OUT STD_LOGIC_VECTOR(13 downto 0);
     ddr_ba : OUT STD_LOGIC_VECTOR(1 downto 0);

     -- Local interface signals
     transrequest, readwriten : IN STD_LOGIC;
     transcomplete : OUT STD_LOGIC;
     ddr_address_in : IN STD_LOGIC_VECTOR(16 downto 0);
     ddr_start_block : IN STD_LOGIC_VECTOR(6 downto 0);
     ddr_end_block : IN STD_LOGIC_VECTOR(6 downto 0);

     rstn : IN std_logic;
     clk : IN std_logic
   );
END;

ARCHITECTURE rtl OF jddr_controller IS

  TYPE DDR_CONTROL IS
    RECORD
      csn     : STD_LOGIC_VECTOR(1 downto 0);
      rasn    : STD_LOGIC;
      casn    : STD_LOGIC;
      wen     : STD_LOGIC;
      address : STD_LOGIC_VECTOR(13 downto 0);
      ba      : STD_LOGIC_VECTOR(1 downto 0);
      clken   : STD_LOGIC_VECTOR(1 downto 0);
    END RECORD;

  PROCEDURE NOP(SIGNAL ctl : OUT DDR_CONTROL) IS
  BEGIN
    ctl.csn <= "11"; ctl.rasn <= '1'; ctl.casn <= '1'; 
    ctl.wen <= '1'; ctl.address <= (others => '-'); 
    ctl.ba <= (others => '-');
  END PROCEDURE;

  PROCEDURE PCH(SIGNAL ctl : OUT DDR_CONTROL) IS
  BEGIN
    ctl.csn <= "00"; ctl.rasn <= '0'; ctl.casn <= '1'; 
    ctl.wen <= '0';
    ctl.address(10) <= '1';
    ctl.address(9 downto 0) <= (others => '-');
    ctl.address(13 downto 11) <= (others => '-');
    ctl.ba <= (others => '-');
  END PROCEDURE; 

  PROCEDURE ARF( SIGNAL ctl : OUT DDR_CONTROL;
                 CONSTANT chips : STD_LOGIC_VECTOR(1 downto 0)) IS
  BEGIN
    ctl.csn <= NOT chips; ctl.rasn <= '0'; ctl.casn <= '0';
    ctl.wen <= '1';
    ctl.address <= (others => '-'); ctl.ba <= (others => '-');
  END PROCEDURE;

  PROCEDURE ACT( SIGNAL ctl : OUT DDR_CONTROL;
                 CONSTANT chips : STD_LOGIC_VECTOR(1 downto 0);
                 CONSTANT active_ba : STD_LOGIC_VECTOR(1 downto 0);
                 CONSTANT active_row : STD_LOGIC_VECTOR(13 downto 0)) IS
  BEGIN
    ctl.csn <= NOT chips; ctl.rasn <= '0'; ctl.casn <= '1';
    ctl.wen <= '1';
    ctl.address <= active_row;  ctl.ba <= active_ba;
  END PROCEDURE;

  PROCEDURE WRITE( SIGNAL ctl : OUT DDR_CONTROL;
                   CONSTANT chips : STD_LOGIC_VECTOR(1 downto 0);
                   CONSTANT active_ba : STD_LOGIC_VECTOR(1 downto 0);
                   CONSTANT columnin : STD_LOGIC_VECTOR(10 downto 0);
                   CONSTANT precharge : STD_LOGIC) IS
  BEGIN
    ctl.csn <= NOT chips; ctl.rasn <= '1'; ctl.casn <= '0'; 
    ctl.wen <= '0';
    -- Ignore the low 3 bits since we are bursting 8 words
    ctl.address <= "00" & columnin(10) & precharge & columnin(9 downto 3) & "000";
    ctl.ba <= active_ba;
  END PROCEDURE;

  PROCEDURE READ( SIGNAL ctl : OUT DDR_CONTROL;
                  CONSTANT chips : STD_LOGIC_VECTOR(1 downto 0);
                  CONSTANT active_ba : STD_LOGIC_VECTOR(1 downto 0);
                  CONSTANT columnin : STD_LOGIC_VECTOR(10 downto 0);
                  CONSTANT precharge : STD_LOGIC) IS
  BEGIN
    ctl.csn <= NOT chips; ctl.rasn <= '1'; ctl.casn <= '0'; 
    ctl.wen <= '1';
    -- Ignore the low 3 bits since we are bursting 8 words
    ctl.address <= "00" & columnin(10) & precharge & columnin(9 downto 3) & "000";
    ctl.ba <= active_ba;
  END PROCEDURE;

  PROCEDURE LMR( SIGNAL ctl : OUT DDR_CONTROL;
                 CONSTANT value : STD_LOGIC_VECTOR(11 downto 0)) IS
  BEGIN
    ctl.csn <= "00"; ctl.rasn <= '0'; ctl.casn <= '0'; 
    ctl.wen <= '0';
    ctl.address(11 downto 0) <= value;
    ctl.address(13 downto 12) <= (others => '-');
    ctl.ba <= "00";
  END PROCEDURE;

  PROCEDURE ELMR( SIGNAL ctl : OUT DDR_CONTROL;
                  CONSTANT value : STD_LOGIC_VECTOR(11 downto 0)) IS
  BEGIN
    ctl.csn <= "00"; ctl.rasn <= '0'; ctl.casn <= '0'; 
    ctl.wen <= '0';
    ctl.address(11 downto 0) <= value;
    ctl.address(13 downto 12) <= (others => '-');
    ctl.ba <= "01";
  END PROCEDURE;

  TYPE states IS (S_POWERUPDELAY, S_STARTUP_NOP1, S_STARTUP_PCH1, 
                  S_STARTUP_ELMR, S_STARTUP_LMR1, S_STARTUP_NOP2, 
                  S_STARTUP_PCH2, S_STARTUP_ARF, S_STARTUP_ARF_WAIT,
                  S_STARTUP_LMR2, S_STARTUP_PCH2_WAIT, S_STARTUP_PCH1_WAIT,
                  S_PRECHARGE_ALL, S_PRECHARGE_WAIT1, S_PRECHARGE_WAIT2,
                  S_IDLE, S_ARF, S_ARF_WAIT, S_ACT, S_ACT_WAIT1, S_ACT_WAIT2,
                  S_READ, S_WRITE, S_WRITE_WAIT, S_READ_WAIT);

  SIGNAL curr_state : states;
  SIGNAL clear_startup_counter, clear_delay_counter : STD_LOGIC;
  SIGNAL startup_delay_finished, nop200_delay : BOOLEAN;
  SIGNAL count_11 : BOOLEAN;
  -- Refresh outstanding counter
  SIGNAL refreshA_outstanding : BOOLEAN;
  SIGNAL refreshB_outstanding : BOOLEAN;
  SIGNAL refreshA_now, refreshB_now : BOOLEAN;
  SIGNAL refreshedA, refreshedB : STD_LOGIC;
  SIGNAL refreshcnten : STD_LOGIC;

  -- Controls which chip is activate in S_ACT
  SIGNAL active_chips : STD_LOGIC_VECTOR(1 downto 0);

  -- Controls for read/write command
  SIGNAL column : STD_LOGIC_VECTOR(9 downto 0);
  SIGNAL precharge : STD_LOGIC;

  -- Column Counter Signals
  SIGNAL column_we : STD_LOGIC;
  SIGNAL column_write, column_done, column_last : BOOLEAN;

  -- Memory interface signals
  SIGNAL doing_wr_delay : STD_LOGIC_VECTOR(2 downto 0);
  SIGNAL l_doing_wr : STD_LOGIC;
  SIGNAL finished : STD_LOGIC;

  -- SRAM Read Delay Chain
  CONSTANT sram_delay : natural := 10;
  SIGNAL sram_we_r : STD_LOGIC_VECTOR(sram_delay-1 downto 0);
  SIGNAL transcomplete_r : STD_LOGIC_VECTOR(sram_delay -1 downto 0);

  -- Write delay chain
  SIGNAL data_wanted_delay : STD_LOGIC_VECTOR(2 downto 0);
  SIGNAL start_write_delay : STD_LOGIC_VECTOR(EXTRA_WRITE_DELAY downto 0);
  SIGNAL column2 : STD_LOGIC_VECTOR(9 downto 0);
  SIGNAL column2_preload : STD_LOGIC;
  SIGNAL column2_done : BOOLEAN;

  -- Decoded DDR Address
  SIGNAL active_row : STD_LOGIC_VECTOR(13 downto 0);
  SIGNAL active_chip : STD_LOGIC;
  SIGNAL active_ba : STD_LOGIC_VECTOR(1 downto 0);

  -- Local control signals
  SIGNAL l_ddr_ctrl : DDR_CONTROL;

BEGIN
  -- Decode ddr_address to individual parts
  active_row <= ddr_address_in(16 downto 3);
  active_chip <= ddr_address_in(2);
  active_ba <= ddr_address_in(1 downto 0);

  -- Doing write delay chain
  l_doing_wr <= '1' WHEN curr_state = S_WRITE OR
                         curr_state = S_WRITE_WAIT 
                    ELSE '0';
  doing_wr <= doing_wr_delay(1) OR doing_wr_delay(0);

  start_write_delay(EXTRA_WRITE_DELAY) <= '1' when curr_state = S_ACT_WAIT2
                                              else '0';
  
  PROCESS(clk)
  BEGIN
    IF rising_edge(clk) THEN
      doing_wr_delay(0) <= doing_wr_delay(1);   -- Hold doing_wr one additional cycle
      doing_wr_delay(1) <= doing_wr_delay(2);
      doing_wr_delay(2) <= l_doing_wr;

      -- Delays the main state machine before going into S_WRITE
      if EXTRA_WRITE_DELAY > 0 then
        start_write_delay(EXTRA_WRITE_DELAY-1 downto 0) <=
          start_write_delay(EXTRA_WRITE_DELAY downto 1);
      end if;
	
      data_wanted_delay(1 downto 0) <= data_wanted_delay(2 downto 1);
      sram_we_r(sram_delay-2 downto 0) <= sram_we_r(sram_delay-1 downto 1);
      transcomplete_r(sram_delay-2 downto 0) <= transcomplete_r(sram_delay-1 downto 1);
    END IF;
  END PROCESS;

  data_wanted <= data_wanted_delay(0);

  sram_we_r(sram_delay-1) <= '1' WHEN curr_state = S_READ or
                                      curr_state = S_READ_WAIT else '0';
  data_available <= sram_we_r(0);

  -- We will need to delay the finished command for reads to insure data 
  -- has arrived.
  transcomplete_r(sram_delay-1) <= finished;
  transcomplete <= transcomplete_r(0);

  active_chips(0) <= active_chip;
  active_chips(1) <= NOT active_chip;

  -- Delay ctrl to ctrl clock domain
  PROCESS (clk)
  BEGIN
     IF falling_edge(clk) THEN
       ddr_clken <= l_ddr_ctrl.clken;
       ddr_csn <= l_ddr_ctrl.csn;
       ddr_rasn <= l_ddr_ctrl.rasn;
       ddr_casn <= l_ddr_ctrl.casn;
       ddr_wen <= l_ddr_ctrl.wen;
       ddr_address <= l_ddr_ctrl.address;
       ddr_ba <= l_ddr_ctrl.ba;
    END IF;
  END PROCESS;

  PROCESS (clk)
  BEGIN
    IF rising_edge(clk) THEN
      CASE (curr_state) IS
        WHEN S_POWERUPDELAY => 
          NOP(l_ddr_ctrl); l_ddr_ctrl.clken <= "00";
        WHEN S_STARTUP_NOP1 => 
          NOP(l_ddr_ctrl); l_ddr_ctrl.clken <= "11";
        WHEN S_STARTUP_PCH1 => PCH(l_ddr_ctrl);
        WHEN S_STARTUP_PCH1_WAIT => NOP(l_ddr_ctrl);
        WHEN S_STARTUP_ELMR => ELMR(l_ddr_ctrl, ELMR_VALUE);
        WHEN S_STARTUP_LMR1 => LMR(l_ddr_ctrl, LMR_VALUE1);
        WHEN S_STARTUP_NOP2 => NOP(l_ddr_ctrl);
        WHEN S_STARTUP_PCH2 => PCH(l_ddr_ctrl);
        WHEN S_STARTUP_PCH2_WAIT => NOP(l_ddr_ctrl);
        WHEN S_STARTUP_ARF  => ARF(l_ddr_ctrl, "11");
        WHEN S_STARTUP_ARF_WAIT  => NOP(l_ddr_ctrl);
        WHEN S_STARTUP_LMR2 => LMR(l_ddr_ctrl, LMR_VALUE2);
        WHEN S_IDLE => NOP(l_ddr_ctrl);
        WHEN S_ARF => ARF(l_ddr_ctrl, "11");
        WHEN S_ARF_WAIT => NOP(l_ddr_ctrl);
        WHEN S_ACT => ACT(l_ddr_ctrl, active_chips, active_ba, active_row);
        WHEN S_ACT_WAIT1 => NOP(l_ddr_ctrl);
        WHEN S_ACT_WAIT2 => NOP(l_ddr_ctrl);
        WHEN S_READ => 
          READ(l_ddr_ctrl,active_chips, active_ba, column & '0', precharge);
        WHEN S_WRITE => 
          WRITE(l_ddr_ctrl, active_chips, active_ba, column & '0', precharge);
        WHEN S_WRITE_WAIT => NOP(l_ddr_ctrl);
        WHEN S_READ_WAIT => NOP(l_ddr_ctrl);
        WHEN S_PRECHARGE_ALL => PCH(l_ddr_ctrl);
        WHEN S_PRECHARGE_WAIT1 =>  NOP(l_ddr_ctrl);
        WHEN S_PRECHARGE_WAIT2 =>  NOP(l_ddr_ctrl);
      END CASE;
    END IF;
  END PROCESS;

  -- Another column counter, used for keeping data_wanted high
  -- for the correct number of cycles during a DDR write.
  process (clk)
  begin
    if rising_edge(clk) then
      if column2_preload = '1' then
        column2 <= (ddr_end_block - ddr_start_block) & "111";
      else
        column2 <= column2 - 1;
      end if;
    end if;
  end process;

  column2_done <= (column2 = 0);


  ddr_counter_inst : jddr_counters
    GENERIC MAP(STARTUPDELAY)
    PORT MAP( column_we, ddr_start_block, ddr_end_block,column_write, 
              column_done, column_last, column, 
              clear_startup_counter, refreshcnten, startup_delay_finished,
              nop200_delay, clear_delay_counter, count_11,
              refreshA_outstanding, refreshB_outstanding,
              refreshA_now, refreshB_now, refreshedA, refreshedB, 
              rstn, clk);

  clear_startup_counter <= '1' WHEN (curr_state = S_STARTUP_LMR1) OR
                          (curr_state = S_STARTUP_ARF_WAIT)
                     ELSE '0';
  clear_delay_counter <= '1' WHEN (curr_state = S_STARTUP_ARF) OR
                                  (curr_state = S_ARF) ELSE '0';
  refreshedA <= '1' WHEN (curr_state = S_ARF) AND refreshA_outstanding ELSE '0';
  refreshedB <= '1' WHEN (curr_state = S_ARF) AND refreshB_outstanding ELSE '0';

  column_we <= '1' WHEN (curr_state = S_ACT_WAIT2) ELSE '0';

  PROCESS (clk,rstn)
  BEGIN
    IF rstn = '0' THEN
      curr_state <= S_POWERUPDELAY;
      refreshcnten <= '0';
      precharge <= '0';
      finished <= '0';
    ELSIF rising_edge(clk) THEN
      CASE (curr_state) IS
        -- Startup sequence states
        WHEN S_POWERUPDELAY =>
          IF startup_delay_finished THEN
            curr_state <= S_STARTUP_NOP1;
          END IF;
        WHEN S_STARTUP_NOP1 => curr_state <= S_STARTUP_PCH1;
        WHEN S_STARTUP_PCH1 => curr_state <= S_STARTUP_PCH1_WAIT;
        WHEN S_STARTUP_PCH1_WAIT =>
          curr_state <= S_STARTUP_ELMR;
        WHEN S_STARTUP_ELMR => curr_state <= S_STARTUP_LMR1;
        WHEN S_STARTUP_LMR1 => 
          curr_state <= S_STARTUP_NOP2;
        WHEN S_STARTUP_NOP2 => 
          IF nop200_delay THEN
            curr_state <= S_STARTUP_PCH2;
          END IF;
        WHEN S_STARTUP_PCH2 =>
          curr_state <= S_STARTUP_PCH2_WAIT;
        WHEN S_STARTUP_PCH2_WAIT =>
          curr_state <= S_STARTUP_ARF;
        WHEN S_STARTUP_ARF =>
          curr_state <= S_STARTUP_ARF_WAIT;
        WHEN S_STARTUP_ARF_WAIT =>
          IF count_11 THEN
            curr_state <= S_STARTUP_LMR2;
          END IF;
        WHEN S_STARTUP_LMR2 =>
          curr_state <= S_IDLE;
          refreshcnten <= '1';  -- enable refresh counter
        -- Controller in run mode
        WHEN S_IDLE =>
          finished <= finished AND transrequest;
          IF (refreshA_outstanding OR refreshB_outstanding) OR
             (transrequest = '1' and finished = '0') THEN
            curr_state <= S_ARF;
          END IF;
        WHEN S_ARF =>
            curr_state <= S_ARF_WAIT;
        WHEN S_ARF_WAIT =>
          IF count_11 THEN
            IF transrequest = '1' and finished = '0' THEN
              curr_state <= S_ACT;
            ELSE
              curr_state <= S_IDLE;
            END IF;
          END IF;
        WHEN S_ACT =>
          curr_state <= S_ACT_WAIT1;
          precharge <= '0';
        WHEN S_ACT_WAIT1 =>
          curr_state <= S_ACT_WAIT2;
	  column2_preload <= '1';
        WHEN S_ACT_WAIT2 =>
	  column2_preload <= '0';
          IF readwriten = '1' THEN
            curr_state <= S_READ;
          ELSE
	    -- Tell user that we want data for writing
	    data_wanted_delay(2) <= '1';
	    -- Delay the actual writing
	    if start_write_delay(0) = '1' then
              curr_state <= S_WRITE;
	    end if;
          END IF;
        WHEN S_PRECHARGE_ALL =>
          curr_state <= S_PRECHARGE_WAIT1;
        WHEN S_PRECHARGE_WAIT1 =>
          curr_state <= S_PRECHARGE_WAIT2;
        WHEN S_PRECHARGE_WAIT2 =>
          curr_state <= S_ARF;
        WHEN S_READ =>
          curr_state <= S_READ_WAIT;
        WHEN S_READ_WAIT =>
          IF column_done THEN
            finished <= '1';
            curr_state <= S_PRECHARGE_ALL;
          ELSIF column_write THEN
            IF column_last THEN
              precharge <= '1';
            END IF;
            curr_state <= S_READ;
          END IF;
        WHEN S_WRITE =>
	  -- Turn off data_wanted signal when
	  -- correct number of cycles have elapsed.
	  if column2_done then
	    data_wanted_delay(2) <= '0';
	  end if;
          curr_state <= S_WRITE_WAIT;
        WHEN S_WRITE_WAIT =>
	  -- Need to check for this signal in both WRITE and WRITE_WAIT
	  if column2_done then
	    data_wanted_delay(2) <= '0';
	  end if;
          IF column_done THEN
            curr_state <= S_IDLE;
            finished <= '1';
          ELSIF column_write THEN
            -- Check if this new write is last and will require auto precharge
            IF column_last THEN
              precharge <= '1';
            END IF;
            curr_state <= S_WRITE;
          END IF;
      END CASE;
    END IF;
  END PROCESS;

END rtl;
