library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

LIBRARY lpm;
USE lpm.lpm_components.all;

ENTITY jddr_counters IS
  GENERIC(STARTUPDELAY : natural);
  PORT(
    -- Column Counter
    column_we : IN STD_LOGIC;
    column_start_block : IN STD_LOGIC_VECTOR(6 downto 0);
    column_end_block : IN STD_LOGIC_VECTOR(6 downto 0);

    column_write : OUT BOOLEAN;
    column_done  : OUT BOOLEAN;
    column_last  : OUT BOOLEAN;
    column       : BUFFER STD_LOGIC_VECTOR(9 downto 0);

    -- Counter 1 signals
    clearstartup : IN STD_LOGIC;
    refreshcnten : IN STD_LOGIC;
    startup_delay : OUT BOOLEAN;
    nop200_delay : OUT BOOLEAN;

    -- Counter 2 signals
    cleardelaycounter : IN STD_LOGIC;
    count_11 : OUT BOOLEAN;

    -- Refresh outstanding counter
    refreshA_outstanding : OUT BOOLEAN;
    refreshB_outstanding : OUT BOOLEAN;
    refreshA_now : OUT BOOLEAN;
    refreshB_now : OUT BOOLEAN;
    refreshedA : IN STD_LOGIC;
    refreshedB : IN STD_LOGIC;
    
    rstn : IN STD_LOGIC;
    clock : IN STD_LOGIC);
END;

ARCHITECTURE rtl OF jddr_counters IS

  COMPONENT lpm_updown3 PORT (
    clock : IN STD_LOGIC;
    updown : IN STD_LOGIC;
    clk_en : IN STD_LOGIC;
    aclr : IN STD_LOGIC;
    q : OUT STD_LOGIC_VECTOR (2 DOWNTO 0));
  END COMPONENT;

  SIGNAL aclr, sclr, refreshclr : STD_LOGIC;
  SIGNAL count1 : STD_LOGIC_VECTOR(15 DOWNTO 0);
  SIGNAL count2 : STD_LOGIC_VECTOR(3 DOWNTO 0);
  SIGNAL count3A, count3B : STD_LOGIC_VECTOR(2 DOWNTO 0);
  SIGNAL columncounter : STD_LOGIC_VECTOR(9 DOWNTO 0);
  SIGNAL updownA, updownB, clk_enA, clk_enB : STD_LOGIC;
BEGIN
  aclr <= NOT rstn;
  sclr <= clearstartup WHEN refreshcnten = '0' ELSE refreshclr;
  refreshclr <= '1' WHEN count1 = CONV_STD_LOGIC_VECTOR(1000, 16) ELSE '0';

  -- Instantiate Column Counter
  PROCESS (clock)
  BEGIN
    IF rising_edge(clock) THEN
      IF column_we = '1' THEN
        columncounter <= column_start_block & "000";
      ELSE
        columncounter <= columncounter + 1;
      END IF;
    END IF;
  END PROCESS;

  column_write <= (columncounter(1 downto 0) = "11");
  column_done  <= (columncounter = column_end_block & "111");
  column_last  <= (columncounter = column_end_block & "011");
  column <= columncounter;

  -- Instantiate refresh and startup counter          
  PROCESS (clock, aclr)
  BEGIN
    IF (aclr = '1') THEN
      count1 <= (others => '0');
    ELSIF rising_edge(clock) THEN
      IF sclr = '1' THEN
        count1 <= (others => '0');
      ELSE
        count1 <= count1 + 1;
      END IF;
    END IF;
  END PROCESS;

  startup_delay <= (count1 = CONV_STD_LOGIC_VECTOR(STARTUPDELAY,16));
  nop200_delay <= (count1 = CONV_STD_LOGIC_VECTOR(200,16));

  -- Instantiate latency counter
  PROCESS (clock, aclr)
  BEGIN
    IF (aclr = '1') THEN
       count2 <= (others => '0');
    ELSIF rising_edge(clock) THEN
      IF cleardelaycounter = '1' THEN 
        count2 <= (others => '0');
      ELSE
        count2 <= count2 + 1;
      END IF;
    END IF;
  END PROCESS;

  count_11 <= (count2 = CONV_STD_LOGIC_VECTOR(11, 4));

  -- Instantiate refresh outstanding counter
  refresh_oustandingA : lpm_counter
    GENERIC MAP (
      lpm_width => 3,
      lpm_type => "LPM_COUNTER", 
      lpm_direction => "UNUSED")
    PORT MAP(
      clk_en => clk_enA,
      aclr => aclr, 
      clock => clock, 
      updown => updownA, 
      q => count3A);

  refresh_oustandingB : lpm_counter
    GENERIC MAP (
      lpm_width => 3,
      lpm_type => "LPM_COUNTER", 
      lpm_direction => "UNUSED")
    PORT MAP(
      clk_en => clk_enB,
      aclr => aclr, 
      clock => clock, 
      updown => updownB, 
      q => count3B);

--  refresh_outstandingA : lpm_updown3
--    PORT MAP( clock, updownA, clk_enA, aclr, count3A );
--  refresh_outstandingB : lpm_updown3
--    PORT MAP( clock, updownB, clk_enB, aclr, count3B );
 
  updownA <= NOT refreshedA; -- Count down if we are refreshing
  updownB <= NOT refreshedB;

  -- Add one to the counters if we have issued a refresh
  -- Unless the ddr controller is indicated it just finished a refresh
  clk_enA <= '1' WHEN ((refreshclr = '1') AND (refreshedA = '0')) OR 
                      ((refreshclr = '0') and (refreshedA = '1'))ELSE '0';
  clk_enB <= '1' WHEN ((refreshclr = '1') AND (refreshedB = '0')) OR 
                      ((refreshclr = '0') and (refreshedB = '1'))ELSE '0';

  -- Signal if we have queued up refreshed
  refreshA_outstanding <= (count3A /= "000");
  refreshB_outstanding <= (count3B /= "000");

  -- Signal if we can't wait any longer for a refresh
  refreshA_now <= (count3A = "111");
  refreshB_now <= (count3B = "111");

END rtl;
